# SPDX-License-Identifier: BSD-2-Clause

import sys

# Requires python 3.7 or greater
if (sys.version_info[0], sys.version_info[1]) < (3, 7):
    sys.exit("Python 3.7 or a more recent version is required.")